import numpy as np
from enum import Enum


class EsriCoordinateType(Enum):
    center = 0
    corner = 1


class EsriAscii:
    def __init__(self,
                 rows,
                 cols,
                 data_array = None,
                 cellsize = 1,
                 nodata_value = -9999,
                 coord_type = EsriCoordinateType.corner):
        if cellsize < 0:
            raise ValueError("cellsize must be greater than 0")
        if type(rows) is not int:
            raise TypeError("rows must be integer")
        if rows < 0:
            raise ValueError("rows must be greater than 0")
        if type(cols) is not int:
            raise TypeError("cols must be integer")
        if cols < 0:
            raise ValueError("cols must be greater than 0")
        if data_array is not None:
            assert((rows, cols) == data_array.shape[0:2])
        self.rows = rows
        self.type = coord_type
        self.cols = cols
        self.xll = 0
        self.yll = 0
        self.cellsize = cellsize
        self.nodata_value = nodata_value
        self.data = data_array if data_array is not None else np.empty((rows, cols))

    def __str__(self):
        return self.to_ascii()

    @property
    def header(self):
        return self.get_header()

    @property
    def sdata(self):
        return self.data_as_string()

    @classmethod
    def from_ndarray(cls,
                     arr,
                     cellsize = 1,
                     nodata_value = -9999,
                     coord_type = EsriCoordinateType.corner):
        """
        Create an Esri raster map from a numpy array

        :param arr: The numpy array
        :param cellsize: Optional Cell Size
        :param nodata_value: Value for no data
        :param coord_type: Should the metadata use corner or center for xll and yll
        :return: The Esri raster map
        """
        if len(arr.shape) == 4:
            arr = arr[..., 0:2]
        if len(arr.shape) == 3:
            arr = np.dot(arr[..., :3], [0.299, 0.587, 0.114])
        ro, co = arr.shape[0:2]
        return cls(ro, co, arr, cellsize=cellsize, nodata_value=nodata_value, coord_type=coord_type)

    def data_as_string(self):
        """Return the data as a string

        :return: A string representing the data
        """
        # return str(self.data)
        ro, co = self.data.shape[0:2]
        out = []
        for r in range(ro):
            # out.append(str(self.data[r]).replace(".", "")[1:-1])
            out.append(' '.join(map(str, self.data[r])))
        return "\n".join(out)

    def get_header(self):
        """Get the Ascii raster header

        :return: The header, as a string
        """
        header = ""
        header += "ncols " + str(self.cols) + "\n"
        header += "nrows " + str(self.rows) + "\n"
        header += "xllcenter " + str(self.xll) + "\n"
        header += "yllcenter " +str(self.yll) + "\n"
        header += "cellsize " +str(self.cellsize) + "\n"
        header += "nodata_value " + str(self.nodata_value) + "\n"
        if self.type == EsriCoordinateType.corner:
            header = header.replace("xllcenter", "xllcorner").replace("yllcenter", "yllcorner")
        return header

    @classmethod
    def from_ascii(cls, ascii_text):
        """Create an EsriAscii object from an ascii text. Data length must match rows and columns
        (length of data = rows * columns)

        :param ascii_text: The text (Should be an Esri Ascii raster formatted text)
        :return: an EsriAscii map
        """
        lines = ascii_text.splitlines()
        assert len(lines) > 6
        _, s_ncols = lines[0].split(" ")
        ncols = int(s_ncols)
        _, s_nrows = lines[1].split(" ")
        nrows = int(s_nrows)
        alignment, s_xll = lines[2].split(" ")
        xll = int(s_xll)
        if alignment[3:] == "center":
            coord = EsriCoordinateType.center
        else:
            coord = EsriCoordinateType.corner
        _, s_yll = lines[3].split(" ")
        yll = int(s_yll)
        _, s_cellsize = lines[4].split(" ")
        cellsize = int(s_cellsize)
        _, s_nodata = lines[5].split(" ")
        nodata = int(s_nodata)
        s_data = " ".join(lines[6:])
        data = np.array(s_data.split(" "), dtype=np.int32).reshape((nrows, ncols))
        ascii_map = cls(rows=nrows,
                        cols=ncols,
                        cellsize=cellsize,
                        data_array=data,
                        nodata_value=nodata,
                        coord_type=coord,
                        )
        ascii_map.xll = int(xll)
        ascii_map.yll = int(yll)
        return ascii_map

    def to_ascii(self):
        """Return the map as text

        :return: The Esri raster as a text
        """
        return self.header + self.sdata

    @classmethod
    def from_file(cls, path):
        with open(path) as fp:
            content = fp.read()
        return cls.from_ascii(content)

    def to_file(self, path):
        """Write the ascii raster to a file.

        :param path: The file to output
        :return: nothing
        """
        with open(path, 'w') as fp:
            fp.write(self.to_ascii())

    def to_array(self):
        """Get the data as a numpy array

        :return: The map data
        """
        return self.data
